from typing import Any, Optional, Union

import asyncio
import logging
from collections.abc import AsyncIterator
from contextlib import suppress
from logging import Logger
from urllib.parse import urlparse

from aiokafka import AIOKafkaConsumer, AIOKafkaProducer, TopicPartition

from rhubarb.backends.base import BaseBackend
from rhubarb.event import Event


class KafkaBackend(BaseBackend):
    def __init__(self, url: str):
        """aiokafka backend

        :param url: the kafka URL to connect to
        :type url: str
        """
        self._servers: list[str] = [urlparse(url).netloc]
        self._channels: set[str] = set()
        self._lock = asyncio.Lock()
        self._consumer_reader_task = None
        self.logger: Logger = logging.getLogger(__name__)

    async def connect(self) -> None:
        """Connects the producer to kafka backend"""
        self.logger.info("Connecting to %s", self._servers)
        self._listen_queue: asyncio.Queue[Union[Event, None]] = asyncio.Queue()
        self._producer = AIOKafkaProducer(bootstrap_servers=self._servers)
        self.logger.info("Starting producer...")
        await self._producer.start()
        self.logger.info("Connected to %s", self._servers)

    async def disconnect(self) -> None:
        """Disconnects and cleans up consumers"""
        self.logger.info("Disconnecting from %s", self._servers)
        await self._producer.stop()
        await self._stop_consumer()
        self.logger.info("Disconnected from %s", self._servers)

    async def _start_consumer(self):
        """Internal start consumer method creating a kafka consumer for the requested channels"""
        self.logger.info("Consuming messages from '%s' channels", self._channels)
        self._consumer = AIOKafkaConsumer(
            *self._channels, bootstrap_servers=self._servers, auto_offset_reset="latest"
        )
        self.logger.info("Starting Kafka consumer...")
        await self._consumer.start()
        self._consumer_reader_task: asyncio.Task[None] = asyncio.create_task(  # type: ignore
            self._reader()
        )
        self.logger.info("Consumer started")

    async def _stop_consumer(self):
        """Stops the consumer task and the AIOKafkaConsumer object"""
        self.logger.info("Stopping consumer...")
        if self._consumer_reader_task:
            self.logger.info("Cancel consumer task...")  # type: ignore
            self._consumer_reader_task.cancel()
            with suppress(asyncio.exceptions.CancelledError):
                await self._consumer_reader_task
            self.logger.info("Stopping Kafka Consumer...")
            await self._consumer.stop()

    async def subscribe(self, channel: str) -> None:
        """Adds the provided channel to the set of ``self.channels`` and restarts the consumer

        :param channel:
        :type channel: str
        """
        self.logger.info("Subscribing to %s", channel)
        async with self._lock:
            await self._stop_consumer()
            self._channels.add(channel)
            await self._start_consumer()
        self.logger.info("Subscribed to %s", channel)

    async def unsubscribe(self, channel: str) -> None:
        """Removes the channel from the ``channels`` and re-creates the consumer

        :param channel: the channel to unsubscribe
        :type channel: str
        """
        self.logger.info("Unsubscribing from %s", channel)
        async with self._lock:
            if channel in self._channels:
                await self._stop_consumer()
                self._channels.remove(channel)
                if self._channels:
                    await self._start_consumer()
                self.logger.info("Unsubscribed from %s", channel)
            else:
                self.logger.warning("Unknown channel %s", channel)

    async def publish(self, channel: str, message: Any) -> None:
        """Using the created ``self._producer`` publish a message to the provided channel

        :param channel: Channel name to use to publish the message to
        :type channel: str
        :param message: Data that is to be published to the channel
        :type message: Any
        """
        self.logger.debug("Publishing message %s to channel %s", message, channel)
        await self._producer.send_and_wait(channel, message.encode("utf8"))

    async def _reader(self) -> None:
        """Read data from the consumer and put events into a queue to be read by the caller."""
        self.logger.debug("Reader started...")
        async for message in self._consumer:
            self.logger.debug(
                "Read message %s from channel: %s",
                message.value.decode("utf8"),
                message.topic,
            )
            event = Event(channel=message.topic, message=message.value.decode("utf8"))
            self._listen_queue.put_nowait(event)

    async def next_event(self) -> Optional[Event]:
        """Return the next event from the queue that was read from all channels"""
        self.logger.debug("Getting next event from kafka...")
        return await self._listen_queue.get()

    async def _get_channel_history(
        self, channel: str, count: int, consumer: AIOKafkaConsumer
    ) -> AsyncIterator[Event]:
        """For a given channel will retrieve the last messages in the stream based starting `count` number of messages back

        :param channel: the channel to retrieve the historic events from
        :type channel: str
        :param count: how many events to reread
        :type count: int
        :param consumer: the consumer constructed by the caller for reading events
        :type consumer: AIOKafkaConsumer

        :return: AsyncIterator of events
        """
        partitions = {
            TopicPartition(channel, part)
            for part in consumer.partitions_for_topic(channel)
        }  # get all the partitions for the topic (channel) which we are subscribing to
        end_offsets = await consumer.end_offsets(partitions)  # get the lastest offset
        for tp, offset_meta in end_offsets.items():
            if offset_meta <= count:  # can't set and offset less than 0
                self.logger.debug("Seeking to the beginning of channel %s", channel)
                await consumer.seek_to_beginning(tp)
            else:
                self.logger.debug(
                    "Seeking channel %s to offset %d", channel, offset_meta - count
                )
                consumer.seek(
                    tp, offset_meta - count
                )  # seek back to reread the stream in order

            while partitions:
                messages = await consumer.getmany(*partitions, timeout_ms=100)

                for tp, batch in messages.items():
                    for message in batch:
                        yield Event(
                            channel=channel, message=message.value.decode("utf8")
                        )
                    partitions.remove(tp)

    async def history(self, channel: str, count: int) -> AsyncIterator[Event]:
        """Used by Rhubarb to retrieve a list of events"""
        if count > 0:
            self.logger.info("Reading the last %d events from '%s'", count, channel)
            consumer = AIOKafkaConsumer(
                channel, bootstrap_servers=self._servers, auto_offset_reset="latest"
            )
            await consumer.start()
            async for event in self._get_channel_history(channel, count, consumer):
                yield event
            await consumer.stop()



from typing import Any, Callable, Optional, Union

import asyncio
import json
import logging
from collections.abc import AsyncGenerator, AsyncIterator
from contextlib import asynccontextmanager, suppress
from logging import Logger
from urllib.parse import urlparse

from .backends.base import BaseBackend
from .event import Event


class Unsubscribed(Exception):
    pass


class UnknownBackend(Exception):
    pass


class SubscribeError(Exception):
    pass


class Rhubarb:
    """An event bus class that abstracts access to different backend queues into a single ``asyncio.Queue``.
    This allows for updating the backend implementation of where events are subscribed from and published to without needing to change the code within each service that uses the queue.

    Has a simple to use interface that can support each combination of single and multi consumer patterns, as well as single and multi producer patterns.

    An example of using the class as a consumer of the event bus

    .. code-block:: python

        async with Rhubarb(EVENT_QUEUE_URL) as events:
            async with events.subscribe(channel="CHATROOM") as subscriber:
                async for event in subscriber:
                    await websocket.send_text(event.message)

    It's also possible to use the same class as a producer of events on the event bus

    .. code-block:: python

        async with Rhubarb(EVENT_QUEUE_URL) as events:
            await events.publish(channel="CHATROOM", message=json.dumps(data))

    It is also possible to parse messages with your own callable:

    .. code-block:: python

        async with Rhubarb(EVENT_QUEUE_URL, serializer=json.loads, deserializer=json.dumps) as events:
            async with events.subscribe(channel="CHATROOM") as subscriber:
                async for event in subscriber:
                    await websocket.send_json(event.message)

    In addition to parsing, it is also possible to publish JSON messages:

    .. code-block:: python

        async with Rhubarb(EVENT_QUEUE_URL, serializer=json.loads, deserializer=json.dumps) as events:
            await events.publish(channel="CHATROOM", message=data)

    """

    def __init__(
        self,
        url: str,
        serializer: Optional[Callable[[Any], str]] = None,
        deserializer: Optional[Callable[[str], Any]] = None,
    ):
        """Constructor for the event bus that configures the event bus
        `_backend`` is intended to be replaced with different backends where required.

        :param url: URL for the backend service
        :type url: str
        :param serializer: URL for the backend service
        :type serializer: Optional[Callable[[Any], str]]
        :param deserializer: URL for the backend service
        :type deserializer: Optional[Callable[[str], Any]]
        """
        self.logger: Logger = logging.getLogger(__name__)
        backend_cls = self.get_backend(url)
        self._backend: BaseBackend = backend_cls(url)
        self._serializer: Optional[Callable[[Any], str]] = serializer
        self._deserializer: Optional[Callable[[str], Any]] = deserializer
        self._subscribers: dict[str, set[asyncio.Queue[Union[Event, None]]]] = {}
        self._lock = asyncio.Lock()
        self._connected = False

    def get_backend(self, url: str) -> BaseBackend:
        parsed_url = urlparse(url)
        self.logger.info("Loading backend for: '%s", parsed_url.scheme)
        if parsed_url.scheme == "redis":
            from rhubarb.backends.redis import RedisBackend

            return RedisBackend
        elif parsed_url.scheme == "kafka":
            from rhubarb.backends.kafka import KafkaBackend

            return KafkaBackend
        elif parsed_url.scheme == "postgres":
            from rhubarb.backends.postgres import AsyncPgBackend

            return AsyncPgBackend
        elif parsed_url.scheme == "memory":
            from rhubarb.backends.memory import MemoryBackend

            return MemoryBackend
        elif parsed_url.scheme == "amqp":
            from rhubarb.backends.rabbitmq import RabbitMQBackend

            return RabbitMQBackend
        else:
            raise UnknownBackend(f'"{parsed_url.scheme}" is not a supported backend!')

    async def _reader(self) -> None:
        """Reads the backend queue and passes events to Subscribers"""
        self.logger.info("Started reading backend events")
        while True:
            event = await self._backend.next_event()
            self.logger.debug("Read new event from backend")
            for queue in self._subscribers.get(
                event.channel, []
            ):  # publish event to all subscribers for this event
                await queue.put(event)

    async def connect(self) -> None:
        """Connect to backend and read Events from the queue"""
        self.logger.info("Connecting to backend")
        async with self._lock:
            if not self._connected:
                await self._backend.connect()
                self._reader_task = asyncio.create_task(self._reader())
                self._connected = True
                self.logger.info("Connected to backend!")
            else:
                self.logger.warning(
                    "Already connected '%s', was 'connect' called more than once?",
                    self._connected,
                )

    async def _close_subscriptions(self) -> None:
        """Close queues for subscribers"""
        self.logger.debug("Closing all subscriptions...")
        for subscribers in self._subscribers.values():
            for queue in subscribers:
                await queue.put(None)

    async def disconnect(self) -> None:
        """Close the running backend reader and ensure that the backend is disconnected"""
        self.logger.info("Disconnecting from backend")
        async with self._lock:
            if self._reader_task.done():
                with suppress(asyncio.exceptions.CancelledError):
                    self._reader_task.result()
            else:
                self._reader_task.cancel()
                with suppress(asyncio.exceptions.CancelledError):
                    await self._reader_task

            await self._close_subscriptions()  # close subscriptions so that subscribers to the event bus stop waiting for events
            await self._backend.disconnect()
            self._connected = False
        self.logger.info("Disonnected from backend")

    async def __aenter__(self) -> "Rhubarb":
        """Allows for the queue to be used in a context manager"""
        await self.connect()
        return self

    async def __aexit__(self, *args: Any, **kwargs: Any) -> None:
        """Disconnect the backend when exiting the context manager"""
        await self.disconnect()

    async def publish(self, channel: str, message: Any) -> None:
        """Publishes a message to the channel provided

        :param channel: channel name
        :type channel: str
        :param message: message to publish to the provided channel name
        :type message: Any
        """
        _message = None
        if self._serializer:  # if a serializer has been provided then call
            self.logger.debug("Serializing message")
            _message = self._serializer(message)
        else:
            _message = message

        self.logger.debug("Publishing to %s message: %s", channel, _message)
        await self._backend.publish(channel, _message)

    async def _get_history(
        self,
        channel: str,
        queue: asyncio.Queue[Union[Event, None]],
        count: Optional[int] = 0,
    ) -> None:
        """Get the last `n` events from the backend

        :param channel: channel name to retrieve the history of
        :type channel: str
        :param queue: the queue to write historical events to
        :type queue: asyncio.Queue
        :param count: number of events to retrieve from the history
        :type count: int
        """
        self.logger.info("Retrieving %s events from channel '%s'", count, channel)
        if count is not None and count > 0:
            async for event in self._backend.history(channel, count=count):
                queue.put_nowait(event)

    async def _subscribe(
        self,
        channel: str,
        queue: asyncio.Queue[Union[Event, None]],
        history: Optional[int] = 0,
    ) -> asyncio.Queue[Union[Event, None]]:
        """Subscribes a queue to the channel and subscribes to the channel in the backend

        Will only subscribe the backend if the channel has not already been used.

        :param channel: channel name to subscribe to
        :type channel: str
        :param queue: new subscribers ``asyncio.Queue``
        :type queue: asyncio.Queue
        """
        self.logger.info(
            "Subscribing to '%s', getting %d historic events", channel, history
        )
        async with self._lock:  # ensure that only one co-routine mutates state at a time
            await self._get_history(channel, queue, count=history)

            if not self._subscribers.get(
                channel
            ):  # if channel doesn't exist or the set is empty then subscribe on the backend
                self.logger.debug(
                    "New subscription, subscribing to '%s' on the backend", channel
                )
                await self._backend.subscribe(channel)
                self._subscribers[channel] = {queue}
            else:
                self.logger.debug(
                    "Already subscribed to '%s', appending queue", channel
                )
                self._subscribers[channel].add(queue)
        self.logger.debug("Current subscribers %s", " ".join(self._subscribers.keys()))
        return queue

    async def _unsubscribe(
        self, channel: str, queue: asyncio.Queue[Union[Event, None]]
    ) -> None:
        """Unsubscribe from the channel

        Will only unsubscribe the backend if the channel is no longer used.

        :param channel: channel name to subscribe to
        :type channel: str
        :param queue: Queue to remove from subscribers
        :type queue: asyncio.Queue
        """
        self.logger.debug("Unsubscribing from %s", channel)
        async with self._lock:  # ensure that only one co-routine mutates state at a time
            if (subscriber := self._subscribers.get(channel)) is not None:
                subscriber.remove(queue)
                if not subscriber:
                    self.logger.debug(
                        "No more subscriptions, unsubscribing from backend."
                    )
                    del self._subscribers[channel]
                    await self._backend.unsubscribe(channel)
        self.logger.debug("Subscribed to %s", " ".join(self._subscribers.keys()))

    async def _group_unsubscribe(
        self,
        channel: str,
        queue: asyncio.Queue[Union[Event, None]],
        group_name: Optional[str],
        consumer_name: Optional[str],
    ) -> None:
        """Handles unsubscribing a single consumer from a group

        Will always call the backend to unsubscribe

        :param channel: channel name to subscribe to
        :type channel: str
        :param queue: Queue to remove from subscribers
        :type queue: asyncio.Queue
        :param group_name: the name of the group this subscriber will join
        :type group_name: str
        :param consumer_name: the unique name in the group for this subscriber
        :type consumer_name: str
        """
        self.logger.info(
            "Unsubscribing from '%s' group: '%s' consumer: '%s'",
            channel,
            group_name,
            consumer_name,
        )
        await self._backend.group_unsubscribe(
            channel=channel, group_name=group_name, consumer_name=consumer_name
        )

    async def _group_subscribe(
        self,
        channel: str,
        queue: asyncio.Queue[Union[Event, None]],
        group_name: Optional[str],
        consumer_name: Optional[str],
    ) -> None:
        """Subscribes a consumer to a group, will request the group to be created in the backend if required.

        Will only subscribe the backend if the channel has not already been used.

        :param channel: channel name to subscribe to
        :type channel: str
        :param queue: new subscribers ``asyncio.Queue``
        :type queue: asyncio.Queue
        """
        self.logger.info(
            "Subscribing to '%s' group: '%s' consumer: '%s'",
            channel,
            group_name,
            consumer_name,
        )
        async with self._lock:  # ensure that only one co-routine mutates state at a time
            await self._backend.group_subscribe(
                channel, group_name, consumer_name, queue
            )

    @asynccontextmanager
    async def subscribe(
        self,
        channel: str,
        history: Optional[int] = 0,
        group_name: Optional[str] = None,
        consumer_name: Optional[str] = None,
    ) -> AsyncIterator["Subscriber"]:
        """A context manager that will yield a subscriber

        :param channel: channel name to subscribe to
        :type channel: str
        :param history: the number of historical events to retrieve, defaults to 0
        :type history: int
        :param group_name: the name of the group this subscriber will join
        :type group_name: str
        :param consumer_name: the unique name in the group for this subscriber
        :type consumer_name: str
        """
        if (group_name or consumer_name) and (
            group_name is None or consumer_name is None
        ):
            raise SubscribeError(f"Must specify both 'group_name' and 'consumer_name'")

        group_subscribe = group_name is not None and consumer_name is not None

        queue: asyncio.Queue[Union[Event, None]] = asyncio.Queue()
        try:
            # branch here to create a different set of group subscribers, keyed on the group and consumer name
            if group_subscribe:
                self.logger.debug(
                    "Group '%s' subscription to '%s' queue: '%s'",
                    group_name,
                    channel,
                    id(queue),
                )
                await self._group_subscribe(channel, queue, group_name, consumer_name)
            else:
                await self._subscribe(channel, queue, history)

            yield Subscriber(queue, self._deserializer)
        finally:
            # handle clean up for group subscriber vs fan out subscriber
            if group_subscribe:
                await self._group_unsubscribe(channel, queue, group_name, consumer_name)
            else:
                await self._unsubscribe(channel, queue)

            self.logger.debug("Closing queue for %s channel", channel)
            await queue.put(None)


class Subscriber:
    def __init__(
        self,
        queue: asyncio.Queue[Union[Event, None]],
        deserializer: Optional[Callable[[str], Any]] = None,
    ) -> None:
        """Wraps an ``asyncio.Queue`` to allow the caller to asynchronously iterate over the queue

        :param queue: `asyncio.Queue`` to will contain events for the caller to iterate over
        :type queue: asyncio.Queue
        """
        self._queue = queue
        self._deserializer = deserializer

    async def __aiter__(self) -> AsyncGenerator[Event, None]:
        """Async iterable allows for iterating over subscribed events"""
        try:
            while event := await self.get():
                yield event
                self._queue.task_done()
        except Unsubscribed:
            pass

    async def get(self) -> Event:
        """Get the next item from the subscribers ``_queue``

        :raises: Unsubscribed when the queue is empty
        :returns: Event
        """
        item = await self._queue.get()
        if item is None:
            raise Unsubscribed()

        if self._deserializer:
            return Event(channel=item.channel, message=self._deserializer(item.message))

        return item



from typing import Any, Optional, Union

import asyncio
import logging
from logging import Logger

import asyncpg

from rhubarb.backends.base import BaseBackend
from rhubarb.event import Event


class AsyncPgBackend(BaseBackend):
    def __init__(self, url: str):
        """Backend using asyncpg

        :param url: URL for the postgres instance
        :type url: str
        """
        self.url = url
        self.logger: Logger = logging.getLogger(__name__)

    async def connect(self) -> None:
        """Connect to the postgres instance using ``self.url``"""
        self.logger.info("Connecting to %s", self.url)
        self._connection = await asyncpg.connect(self.url)
        self._listen_queue: asyncio.Queue[Union[Event, None]] = asyncio.Queue()
        self._channels: set[str] = set()
        self.logger.info("Connected to %s", self.url)

    async def disconnect(self) -> None:
        """Disconnect from the postgres instance"""
        self.logger.info("Disconnecting from %s", self.url)
        await self._connection.close()
        self.logger.info("Disconnected from %s", self.url)

    async def subscribe(self, channel: str) -> None:
        """Subscribe to the named channel

        :param channel: the channel to subscribe to
        :type channel: str
        """
        self.logger.info("Subscribing to '%s'", channel)
        await self._connection.add_listener(channel, self._listener)
        self._channels.add(channel)
        self.logger.info("Subscribed to '%s'", channel)

    async def unsubscribe(self, channel: str) -> None:
        """Unsubscribe from the named channel

        :param channel: the channel to unsubscribe to
        :type channel: str
        """
        self.logger.info("Unsubscribing from '%s'", channel)
        if channel in self._channels:
            await self._connection.remove_listener(channel, self._listener)
            self._channels.remove(channel)
        else:
            self.logger.warning("Unknown channel '%s'", channel)

        self.logger.info("Unsubscribed from '%s'", channel)

    async def publish(self, channel: str, message: str) -> None:
        """Publish the passed ``message`` to the provided ``channel``

        :param channel: the channel to publish the ``message`` to.
        :type channel: str
        :param message: the message to publish
        :type message: str
        """
        self.logger.debug("Publishing %s to '%s'", message, channel)
        await self._connection.execute(
            "SELECT pg_notify($1, $2);", channel, str(message)
        )

    def _listener(self, *args: Any) -> None:
        """listener callback for the asyncpg `add_listener` call"""
        _, __, channel, payload = args
        self.logger.debug(
            "Read message %s from channel: %s",
            payload,
            channel,
        )
        event = Event(channel=channel, message=payload)
        self._listen_queue.put_nowait(event)

    async def next_event(self) -> Optional[Event]:
        """Get the next event from the queue"""
        return await self._listen_queue.get()



from typing import Any, Union

import asyncio
import logging
from collections import namedtuple
from collections.abc import AsyncIterator
from contextlib import suppress
from logging import Logger

from redis import asyncio as aioredis

from rhubarb.backends.base import BaseBackend
from rhubarb.event import Event

GroupConsumer = namedtuple("GroupConsumer", "group consumer")


class RedisBackend(BaseBackend):
    def __init__(self, url: str):
        """Backend for the message queue using asyncio ``redis`` to interface with redis

        :param url: The URL of the redis instance
        :type url: str
        """
        self.url: str = url
        self._channels: dict[str, asyncio.Task[None]] = {}  # the channels subscribed to
        self._group_readers: dict[
            str, dict[GroupConsumer, asyncio.Task[None]]
        ] = {}  # hanndles many group readers
        self._channel_latest_id: dict[str, int] = {}
        self.logger: Logger = logging.getLogger(__name__)

    async def connect(self) -> None:
        """Connection to the configured Redis URL using ``aioredis``.
        Execute a `ping` to check if the connection is valid.
        """
        self.logger.info("Connecting to '%s'", self.url)
        self._listen_queue: asyncio.Queue[Union[Event, None]] = asyncio.Queue()
        self._redis = aioredis.from_url(  # type: ignore
            self.url, encoding="utf-8", decode_responses=True
        )
        await self._redis.ping()  # causes aioredis to connect immediately
        self.logger.info("Connected to '%s'", self.url)

    async def disconnect(self) -> None:
        """Gracefully close the queue and end the tasks, the redis connection gracefully closes itself during garbage collection."""
        self.logger.info("Disconnecting from %s", self.url)
        for channel in list(self._channels.keys()):
            await self.unsubscribe(channel)

        for channel in list(self._group_readers.keys()):
            for consumer in list(self._group_readers[channel].keys()):
                await self.group_unsubscribe(
                    channel, consumer.group, consumer.consumer
                )  # pragma: no cover

        self.logger.info("Disconnected from %s", self.url)

    async def subscribe(self, channel: str) -> None:
        """Subscribe to the passed channel

        :param channel: channel name to subscribe to
        :type channel: str
        """
        self.logger.info("Subscribing to '%s'", channel)
        if channel in self._channels:
            self.logger.warning("Already subscribed to '%s'", channel)
        else:
            self._channels[channel] = asyncio.create_task(self._reader(channel))
        self.logger.info("Subscribed to '%s'", channel)

    async def unsubscribe(self, channel: str) -> None:
        """Unsubscribe from the passed channel

        :param channel: channel name to subscribe to
        :type channel: str
        """
        self.logger.info("Unsubscribing from '%s'", channel)
        if (subscription := self._channels.get(channel)) is not None:
            self.logger.debug("Cancelling task for '%s' channel", channel)
            self.logger.debug("Task %s", subscription)
            subscription.cancel()
            with suppress(asyncio.CancelledError):
                self.logger.debug("Waiting for '%s' reader task to cancel...", channel)
                del self._channels[channel]  # signal for the reader to close
                await subscription  # wait for the reader to close
        else:
            self.logger.warning("Unknown channel '%s'", channel)

    async def publish(self, channel: str, message: Any) -> None:
        """Using the configured redis connection to publish a message to the provided channel

        :param channel: Channel name to use to publish the message to
        :type channel: str
        :param message: Data that is to be published to the channel
        :type message: Any
        """
        fields = {"message": message}
        self.logger.debug("Publishing message '%s' to channel '%s'", message, channel)
        await self._redis.xadd(channel, fields)

    async def _reader(self, channel: str) -> None:
        """Read data from the passed channel object and put events into a queue to be read by the caller.

        :param channel: The channel subscribed to
        :type channel: str
        """
        self.logger.debug("Reader starting for '%s' channel", channel)
        streams: dict[str, int] = {channel: 0}
        if latest_message_id := self._channel_latest_id.get(
            channel
        ):  # history was called before
            streams[channel] = latest_message_id
        else:
            results = await self._redis.xrevrange(
                channel, count=1
            )  # read the last events from the channel
            for message_id, values in reversed(results):
                streams[channel] = message_id

        self.logger.debug(
            "Latest stream id '%s' for '%s' channel", streams[channel], channel
        )

        while True:
            response = await self._redis.xread(streams, block=100, count=1)
            self.logger.debug(
                "Read %d messages for channel '%s'", len(response), channel
            )
            for stream_name, messages in response:
                for message_id, values in messages:
                    self.logger.debug(
                        "Read message %s from channel '%s', Updating latest stream id '%s'",
                        values["message"],
                        stream_name,
                        message_id,
                    )
                    streams[stream_name] = message_id
                    event = Event(channel=stream_name, message=values["message"])
                    self._listen_queue.put_nowait(event)
            if (
                channel not in self._channels
            ):  # check if the channel has unsubscribed and break out
                break

    async def next_event(self) -> Union[Event, None]:
        """Return the next event from the queue that was read from all channels"""
        self.logger.debug("Getting next event from redis...")
        return await self._listen_queue.get()

    async def history(self, channel: str, count: int = 0) -> AsyncIterator[Event]:
        """Optionally get a history of the last `n` events

        :return: A list of the last events
        :type: List
        """
        if count > 0:
            self.logger.info("Reading the last %s events from '%s'", count, channel)
            results = await self._redis.xrevrange(
                channel, count=count
            )  # read the last `n` events from the channel
            for message_id, values in reversed(results):
                yield Event(channel=channel, message=values["message"])
                self._channel_latest_id[channel] = message_id

    async def _group_reader(
        self,
        channel: str,
        group_consumer: GroupConsumer,
        queue: asyncio.Queue[Union[Event, None]],
    ) -> None:
        """A group reader will take a channel and group_consumer to consume messages that are put onto the queue
        allowing for an at-most-once delivery of messages for a group of subscribers.

        :param channel: The channel subscribed to
        :type channel: str
        :param group_consumer: The channel subscribed to
        :type group_consumer: GroupConsumer
        :param queue: The queue that the consumer is reading from
        :type queue: asyncio.Queue
        """
        self.logger.info(
            "Group Reader starting for channel '%s' group: '%s' consumer name: '%s'",
            channel,
            group_consumer.group,
            group_consumer.consumer,
        )
        streams: dict[str, str] = {channel: ">"}
        while True:
            response = await self._redis.xreadgroup(
                group_consumer.group, group_consumer.consumer, streams=streams, count=1
            )
            for stream_name, messages in response:
                self.logger.debug(
                    "Read %d messages for channel '%s' group '%s' consumer '%s'",
                    len(messages),
                    stream_name,
                    group_consumer.group,
                    group_consumer.consumer,
                )
                for message_id, values in messages:
                    self.logger.debug(
                        "Read message %s from channel '%s' Group '%s'  Consumer '%s', stream id '%s'",
                        values["message"],
                        stream_name,
                        group_consumer.group,
                        group_consumer.consumer,
                        message_id,
                    )
                    event = Event(channel=stream_name, message=values["message"])
                    queue.put_nowait(event)
                    self.logger.debug(
                        "Stream '%s' Consumer '%s' Group '%s' Waiting for processing... %s",
                        stream_name,
                        group_consumer.consumer,
                        group_consumer.group,
                        id(queue),
                    )
                    await queue.join()
                    await self._redis.xack(
                        stream_name, group_consumer.group, message_id
                    )  # acknowledge the message as read after the caller has processed the event
                    self.logger.debug(
                        "Stream '%s' Consumer '%s' Group '%s' Queue finished... %s",
                        stream_name,
                        group_consumer.consumer,
                        group_consumer.group,
                        id(queue),
                    )
            if group_consumer not in self._group_readers.get(  # type: ignore
                channel
            ):  # check if the group consumer has unsubscribed and break out
                break  # pragma: no cover
            else:
                await asyncio.sleep(0)

    async def group_subscribe(
        self,
        channel: str,
        group_name: str,
        consumer_name: str,
        queue: asyncio.Queue[Union[Event, None]],
    ) -> None:
        """Called to subscribe to a channel as part of a consumer (``consumer_name``) within a group (``groupd_name``)

        :param channel: name of the channel in the queue to subscribe to
        :type channel: str
        :param group_name: the name of the group this subscriber will join
        :type group_name: str
        :param consumer_name: the unique name in the group for this subscriber
        :type consumer_name: str
        """
        self.logger.info(
            "Subscribing to '%s' with group '%s' consumer '%s'",
            channel,
            group_name,
            consumer_name,
        )
        try:
            await self._redis.xgroup_create(
                name=channel, groupname=group_name, mkstream=True
            )
        except aioredis.ResponseError as e:  # type: ignore
            if str(e) != "BUSYGROUP Consumer Group name already exists":
                raise  # pragma: no cover

        self.logger.info(
            "Creating reader task on channel '%s' for group '%s' consumer '%s'",
            channel,
            group_name,
            consumer_name,
        )
        if channel not in self._group_readers:
            self._group_readers[channel] = {}

        group_consumer = GroupConsumer(group_name, consumer_name)

        if group_consumer in self._group_readers[channel]:
            self.logger.warning(
                "Consumer %s is already reading in group %s", consumer_name, group_name
            )  # pragma: no cover

        self._group_readers[channel][group_consumer] = asyncio.create_task(
            self._group_reader(channel, group_consumer, queue)
        )
        self.logger.info(
            "Subscribed to '%s' in group '%s' as consumer '%s'",
            channel,
            group_name,
            consumer_name,
        )

    async def group_unsubscribe(
        self, channel: str, group_name: str, consumer_name: str
    ) -> None:
        """Called to unsubscribe from a channel based on the group name and consumer name

        :param channel: name of the channel in the queue to subscribe to
        :type channel: str
        :param group_name: the name of the group this subscriber will join
        :type group_name: str
        :param consumer_name: the unique name in the group for this subscriber
        :type consumer_name: str
        """
        self.logger.info("Unsubscribing from '%s'", channel)
        if (channel := self._group_readers.get(channel)) is not None:  # type: ignore
            self.logger.debug(
                "Cancelling task for '%s' channel group '%s' consumer '%s'",
                channel,
                group_name,
                consumer_name,
            )
            group_consumer = GroupConsumer(group_name, consumer_name)
            if (task := channel.get(group_consumer)) is not None:  # type: ignore
                self.logger.debug("Task %s", task)
                task.cancel()
                with suppress(asyncio.CancelledError):
                    self.logger.debug("Waiting for '%s' reader task to cancel...", task)
                    del channel[group_consumer]  # type: ignore
                    await task  # wait for the reader to close
            else:
                self.logger.warning(
                    "Unknown group '%s' consumer '%s' combination for channel '%s'",
                    group_name,
                    consumer_name,
                    channel,
                )
        else:
            self.logger.warning("Unknown channel '%s'", channel)
